# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from typing import List, Optional

from fhirtordf.rdfsupport.namespaces import FHIR
from rdflib import URIRef, Graph, BNode, RDF, XSD
from rdflib.term import Node, Literal

from i2fhirb2.fhir.fhirencountermapping import FHIREncounterMapping
from i2fhirb2.fhir.fhirpatientdimension import FHIRPatientDimension
from i2fhirb2.fhir.fhirpatientmapping import FHIRPatientMapping
from i2fhirb2.fhir.fhirproviderdimension import FHIRProviderDimension
from i2fhirb2.fhir.fhirspecific import concept_code, composite_uri, instance_is_primitive
from i2fhirb2.fhir.fhirvisitdimension import FHIRVisitDimension
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFact, ObservationFactKey, valuetype_blob, \
    valuetype_text, valuetype_date, valuetype_number
from i2fhirb2.sqlsupport.dynobject import DynElements

# TODO: Same patient_ide/patient_ide_src + visit_ide / visit_ide_src == duplicate


def literal_val(val: Literal) -> str:
    return '"{}"'.format(str(val).replace('\n', '\\n').replace(r'\t', '\\t'))


def boolean_val(val: Literal) -> str:
    return val.value


def date_val(val: Literal) -> datetime:
    return val.value


def datetime_val(val: Literal) -> datetime:
    return datetime.strptime(val.value, '%Y-%m-%dT%H:%M:%SZ')


def decimal_val(val: Literal) -> float:
    return float(val.value)


def gyear_val(val: Literal) -> datetime:
    return datetime.strptime(val.value, '%Y')


def gYearMonth_val(val: Literal) -> datetime:
    return datetime.strptime(val.value, '%Y-%m')


def time_val(val: Literal) -> datetime:
    return datetime.strptime(val.value, '%H:%M:%S')


# Conversion table from XSD data type to corresponding i2b2 field
literal_conversions = {
    XSD.base64Binary: (literal_val, valuetype_blob),
    XSD.boolean: (boolean_val, valuetype_text),
    XSD.date: (date_val, valuetype_date),
    XSD.dateTime: (datetime_val, valuetype_date),
    XSD.decimal: (decimal_val, valuetype_number),
    XSD.gYear: (gyear_val, valuetype_date),
    XSD.gYearMonth: (gYearMonth_val, valuetype_date),
    XSD.integer: (decimal_val, valuetype_number),
    XSD.nonNegativeInteger: (decimal_val, valuetype_number),
    XSD.positiveInteger: (decimal_val, valuetype_number),
    XSD.time: (time_val, valuetype_date)}


class FHIRObservationFact(ObservationFact):
    _t = DynElements(ObservationFact)

    def __init__(self, g: Graph, ofk: ObservationFactKey, concept: URIRef, modifier: Optional[URIRef], obj: Node,
                 instance_num: Optional[int] = None) -> None:
        """
        Construct an observation fact entry
        :param g: Graph containing information about the object
        :param ofk: Observation fact key - patient, encounter, etc
        :param concept: concept URI
        :param modifier: modifier URI (if any)
        :param obj: object of concept_code
        :param instance_num: instance identifier
        """
        super().__init__(ofk, concept_code(concept))
        if modifier is not None:
            self._modifier_cd = concept_code(modifier)
        self._instance_num = instance_num
        self.fhir_primitive(g, obj)

    def fhir_primitive(self, g: Graph, obj: Optional[Node]) -> None:
        assert(instance_is_primitive(g, obj))
        val = g.value(obj, FHIR.value, any=False)
        # TODO: remove the conversions that aren't needed (toPython does the same thing)
        if val.datatype in literal_conversions:
            f, t = literal_conversions[val.datatype]
            if t == valuetype_text:
                self._tval_char = val.toPython()
            elif t == valuetype_number:
                self._tval_char = 'E'
                self._nval_num = val.toPython()
            elif t == valuetype_blob:
                self._observation_blob = val.toPython()
            elif t == valuetype_date:
                dt = val.toPython()
                self._date_val(dt)
            else:
                self._tval_char = val.toPython()
            self._valtype_cd = t.code
        else:
            self._tval_char = val.toPython()
            self._valtype_cd = valuetype_text.code


class FHIRObservationFactFactory:
    """
    FHIRObservationFactFactory takes an RDF graph and generates a set of observation_fact, patient_dimension,
    visit_dimension, provider_dimension, patient_mapping and encounter_mapping entries
    """

    # TODO: FHIR.Observation.referenceRange predicates are a temporary fix to issue 7.  Get a real fix in!
    special_processing_list = {RDF.type: None, FHIR.nodeRole: None, FHIR.index: None, FHIR.link: None,
                               FHIR.Observation.referenceRange: None}

    def __init__(self, g: Graph, ofk: ObservationFactKey, subject: Optional[URIRef]) -> None:
        self.g = g
        self.ofk = ofk
        self.observation_facts = []         # type: List[FHIRObservationFact]
        self.patient_dimensions = []        # type: List[FHIRPatientDimension]
        self.provider_dimensions = []       # type: List[FHIRProviderDimension]
        self.visit_dimensions = []          # type: List[FHIRVisitDimension]
        self.patient_mappings = []          # type: List[FHIRPatientMapping]
        self.encounter_mappings = []        # type: List[FHIREncounterMapping]
        self._inst_num = 0

        # TODO: look at DomainResource embedded entries (claim-example-oral-bridge).  Perhaps we should change the
        #       RDF generator to add type arcs to all resources?
        for s in {subject} if subject else self.g.subjects(FHIR.nodeRole, FHIR.treeRoot):
            self.observation_facts += self.generate_facts(s)

    def generate_facts(self, subject: URIRef) -> List[FHIRObservationFact]:
        """
        Generate an observation fact entry for each predicate object associated with subject.  If the predicate
        is not primitive (i.e. it has a non-value BNode as a target), generate a series of modifiers for each element
        in the BNode.  Note -- this generation sorts the inputs because the order is indirectly preserved in the output
        instance numbers.
        :param subject: Subject of the observation (Patient)
        :return:
        """
        rval = []               # type: List[FHIRObservationFact]

        for conc, obj in sorted(self.g.predicate_objects(subject)):
            if conc not in self.special_processing_list:
                if instance_is_primitive(self.g, obj):
                    rval.append(FHIRObservationFact(self.g, self.ofk, conc, None, obj))
                else:
                    rval += self.generate_modifiers(subject, conc, obj)
        return rval

    def generate_modifiers(self, subject: URIRef, concept: URIRef, obj: BNode,
                           parent_inst_num: int=0, same_subject: bool=False) -> List[FHIRObservationFact]:
        """
        Emit any modifiers for subject/obj.  If there is fhir:index field, this is an indication that this set
        of modifiers can occur multiple times and, as such, must be clustered.  Noting that, at the moment, we
        can only do one level of clustering, we do not generate unique instance numbers for singleton elements.
        :param subject: Resource subject
        :param concept: Concept identifier (first level entry)
        :param obj: inner cluster (usually a BNODE)
        :param parent_inst_num: instance number passed in recursive call
        :param same_subject: True means same parent -- don't revisit instance number
        :return:
        """
        rval = []               # type: List[FHIRObservationFact]
        fhir_index = self.g.value(obj, FHIR.index, any=False) if not same_subject else None
        pred_obj_list = \
            sorted([(p, o) for p, o in self.g.predicate_objects(obj) if p not in self.special_processing_list])

        if fhir_index is not None and len(pred_obj_list):
            self._inst_num += 1
            inst_num = self._inst_num
        else:
            inst_num = parent_inst_num

        for modifier, modifier_object in pred_obj_list:
            if instance_is_primitive(self.g, modifier_object):
                mod_inst_idx = self.g.value(modifier_object, FHIR.index, any=False)
                if mod_inst_idx is not None and len(pred_obj_list) > 1:
                    self._inst_num += 1
                    inst_num = self._inst_num
                obsfact = FHIRObservationFact(self.g, self.ofk, concept, modifier, modifier_object, inst_num)
                rval.append(obsfact)
            else:
                rval += self.generate_modifiers(subject,
                                                concept if inst_num else composite_uri(concept, modifier),
                                                modifier_object,
                                                inst_num,
                                                same_subject=True)
        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import sys
from argparse import Namespace
from datetime import datetime
from pathlib import Path
from typing import List, Dict, Any, Union
from urllib import request
from urllib.error import HTTPError

from fhirtordf.fhir.fhirmetavoc import FHIRMetaVoc

from i2fhirb2.i2b2model.metadata.commondimension import CommonDimension
from i2fhirb2.i2b2model.metadata.i2b2conceptdimension import ConceptDimension
from i2fhirb2.i2b2model.metadata.i2b2conceptdimension import ConceptDimensionRoot
from rdflib import Graph
from sqlalchemy import delete, Table, update

from i2fhirb2 import __version__
from i2fhirb2.fhir.fhirontologytable import FHIROntologyTable
from i2fhirb2.fhir.fhirspecific import FHIR
from i2fhirb2.i2b2model.metadata.i2b2modifierdimension import ModifierDimension
from i2fhirb2.i2b2model.metadata.i2b2ontology import OntologyEntry, OntologyRoot
from i2fhirb2.i2b2model.metadata.i2b2tableaccess import TableAccess
from i2fhirb2.i2b2model.shared.tablenames import i2b2tablenames
from i2fhirb2.sqlsupport.dbconnection import add_connection_args, process_parsed_args, decode_file_args, FileAwareParser
from i2fhirb2.sqlsupport.i2b2tables import I2B2Tables, change_column_length
from i2fhirb2.tsv_support.tsvwriter import write_tsv

Default_Metavoc_URI = "http://build.fhir.org/"
Default_Sourcesystem_Code = 'FHIR STU3'
Default_Base = 'FHIR'
Default_Path_Base = '\\{}\\'.format(Default_Base)

DIMENSION_LIST = Union[List[ConceptDimension], List[ModifierDimension], List[OntologyEntry]]


def pluralize(cnt: int, base: Any) -> str:
    """
    Pluralize base based on the number in cnt
    :param cnt: number of elements
    :param base: base word
    :return: plural
    """
    return str(base) + ('s' if cnt != 1 else '')


def initialize_table_defaults(g: Graph, opts: Namespace) -> None:
    """
    Initialize all of the table defaults
    :param g: Graph containing fhir.ttl
    :param opts: input options
    """
    # TODO: This is kind of messy -- we've refactored so we should be able to consolidate this
    CommonDimension.graph = g

    ConceptDimension._clear()
    ConceptDimension.sourcesystem_cd = opts.sourcesystem
    ConceptDimension.update_date = opts.updatedate

    ConceptDimensionRoot._clear()
    ConceptDimensionRoot.sourcesystem_cd = opts.sourcesystem
    ConceptDimensionRoot.update_date = opts.updatedate

    ModifierDimension._clear()
    ModifierDimension.sourcesystem_cd = opts.sourcesystem
    ModifierDimension.update_date = opts.updatedate

    OntologyEntry._clear()
    OntologyEntry.sourcesystem_cd = opts.sourcesystem
    OntologyEntry.update_date = opts.updatedate

    OntologyRoot._clear()
    OntologyRoot.sourcesystem_cd = opts.sourcesystem
    OntologyRoot.update_date = opts.updatedate


def generate_i2b2_files(g: Graph, opts: Namespace) -> bool:
    """
    Generate the requested i2b2 files as listed in opts
    :param g: Graph of the data to be generated
    :param opts: supplied options
    :return: success indicator
    """
    if not opts.table or opts.table in (i2b2tablenames.concept_dimension,
                                        i2b2tablenames.concept_dimension,
                                        i2b2tablenames.ontology_table):
        resource = FHIR[opts.resource] if opts.resource else None
        initialize_table_defaults(g, opts)
        dimset = FHIROntologyTable(g, name_base=opts.base).dimension_list(resource)

        return \
            output_table_access(opts) and \
            output_concept_dimension(opts, dimset.concept_dimension) and \
            output_modifier_dimension(opts, dimset.modifier_dimension) and \
            output_ontology(opts, dimset.ontology_dimension)


def output_table_access(opts: Namespace) -> bool:
    """
    Generate the table access entries if requested
    :param opts: user options
    :return: True if success or generation is not needed
    """
    if not opts.table or opts.table == i2b2tablenames.table_access:
        table_access = TableAccess()
        if opts.outdir:
            return write_tsv(opts.outdir, 'table_access', table_access._header(), [table_access])
        else:
            return update_table_access_table(opts, opts.tables.table_access, [table_access._freeze()])
    else:
        return True


def update_table_access_table(opts: Namespace, table: Table, records: List[Dict[str, Any]]) -> bool:
    ndel = opts.tables.ont_connection.execute(delete(table).where(table.c.c_table_cd == Default_Base)).rowcount
    if ndel > 0:
        print("{} {} {} deleted".format(ndel, table, pluralize(ndel, "record")))
    nins = opts.tables.ont_connection.execute(table.insert(), records).rowcount
    print("{} {} {} inserted".format(nins, table, pluralize(nins, "record")))
    return True


def output_concept_dimension(opts: Namespace, output: List[ConceptDimension]) -> bool:
    """
    Generate the concept dimension output if required
    :param opts: input options
    :param output: output list
    :return: success indicator
    """
    table_name = i2b2tablenames.concept_dimension
    if not opts.table or opts.table == table_name:
        if opts.outdir:
            return write_tsv(opts.outdir, table_name, ConceptDimension._header(), output)
        else:
            table = opts.tables.concept_dimension
            change_column_length(table, table.c.concept_cd, 200, opts.tables.crc_engine)
            return update_dimension_table(output, opts, table, 'concept_path', [opts.base])
    else:
        return True


def output_modifier_dimension(opts: Namespace, output: List[ModifierDimension]) -> bool:
    """
    Generate the modifier dimension output if required
    :param opts: input options
    :param output: output list
    :return: success indicator
    """
    table_name = i2b2tablenames.modifier_dimension
    if not opts.table or opts.table == table_name:
        if opts.outdir:
            return write_tsv(opts.outdir, table_name, ModifierDimension._header(), output)
        else:
            table = opts.tables.modifier_dimension
            change_column_length(table, table.c.modifier_cd, 200, opts.tables.crc_engine)
            return update_dimension_table(output, opts, table, 'modifier_path', [opts.base])
    else:
        return True


def output_ontology(opts: Namespace, output: List[OntologyEntry]) -> bool:
    """
    Generate the i2b2 ontology output if required
    :param opts: input options
    :param output: set of ontology entries
    :return: success indicator
    """
    table_name = i2b2tablenames.ontology_table
    if not opts.table or opts.table == table_name:
        if opts.outdir:
            return write_tsv(opts.outdir, table_name, OntologyEntry._header(), output)
        else:
            table = opts.tables.ontology_table
            change_column_length(table, table.c.c_basecode, 200, opts.tables.ont_engine)
            # MedicationStatement is 1547 long
            change_column_length(table, table.c.c_tooltip, 1600, opts.tables.ont_engine)
            return update_dimension_table(output, opts, table, 'c_basecode', [opts.base.replace('\\', '') + ':', 'W5'])
    else:
        return True


def update_dimension_table(output: DIMENSION_LIST, opts: Namespace, table: Table, table_key: str, key_match: List[str])\
        -> bool:
    """
    Update the supplied dimension table, removing all existing records
    :param output: list of dimension entries
    :param opts: input options
    :param table: table to be updated
    :param table_key: key to use for removing existing entries
    :param key_match: list of key roots (delete all keys that start with this)
    :return: Success indicator
    """
    for km in key_match:
        q = delete(table).where(table.c[table_key].startswith(km.replace('\\', '\\\\')))
        ndel = opts.tables.crc_connection.execute(q).rowcount
        if ndel > 0:
            print("{} {} {} deleted".format(ndel, table, pluralize(ndel, "record")))
    nins = opts.tables.crc_connection.execute(table.insert(), [e._freeze() for e in output]).rowcount
    print("{} {} {} inserted".format(nins, table, pluralize(nins, "record")))
    return True


def load_fhir_ontology(opts: Namespace) -> Graph:
    """
    Load the fhir specification ontology and w5
    :param opts: User options
    :return: Graph containing the ontology
    """
    print("Loading fhir.ttl")
    fmv = FHIRMetaVoc(os.path.join(opts.metadatavoc, 'fhir.ttl'))
    print(" (cached)" if fmv.from_cache else "(from disc)")
    print("loading w5.ttl")
    fmv.g.load(os.path.join(opts.metadatavoc, 'w5.ttl'), format="turtle")
    print(" done\n")
    return fmv.g


def test_configuration(opts: Namespace) -> bool:
    """ Test the configuration, making sure that the tsv files exist, the database can be opened and that we have
    write access to at least some of the tables
    """
    def test_ttl_file(name: str) -> bool:
        filename = opts.metadatavoc + name
        f = Path(opts.metadatavoc + name)
        if '://' in str(filename):
            try:
                request.urlopen(filename)
            except HTTPError as e_:
                return error("{}: {}".format(filename, e_))
            print('\tURL: {} is valid'.format(filename))
        elif f.is_file():
            print("\tFile: {} exists".format(filename))
        else:
            return error("'{}' file not found in {}".format(name, opts.metadatavoc))
        return True

    def test_tn(name: str, i2b2tables: I2B2Tables) -> bool:
        if name in [k for k, _ in i2b2tables._tables()]:
            print("\tTable {} exists".format(tn))
            return True
        else:
            return error("unable to locate table '{}'".format(tn))

    def error(msg: str) -> bool:
        print("  Error: -> {}".format(msg))
        return False

    print("Validating input files")
    success = test_ttl_file('fhir.ttl')
    if not test_ttl_file('w5.ttl'):
        success = False

    print("Validating sql connection")
    tables = None
    try:
        tables = I2B2Tables(opts)
    except Exception as e:
        print(str(e))

    if not tables:
        success = error("Unable to open SQL database")
        ont_url, crc_url = I2B2Tables._db_urls(opts)
        if ont_url == crc_url:
            print("\tURL: {}".format(ont_url))
        else:
            print("\tOntoogy URL: {}".format(ont_url))
            print("\tCRC URL: {}".format(crc_url))
    else:
        print("\tConnection validated")
        print("Validating target tables")
        for tn in i2b2tablenames.all_tables():
            if not test_tn(i2b2tablenames.phys_name(tn), tables):
                success = False

    if success:
        print("Testing write access")
        row_count = 0
        try:
            row_count = tables.ont_connection.execute(update(tables.table_access)
                                                      .where(tables.table_access.c.c_hlevel == 0)
                                                      .values(c_hlevel=0)).rowcount
        except Exception as e:
            if "permission" not in str(e) or "denied" not in str(e):
                print(str(e))
        if row_count > 0:
            print("\t{} rows updated in table_access table ".format(row_count))
        else:
            success = error("Write permission denied for table_access table")

    return success


def create_parser() -> FileAwareParser:
    """
    Create a command line parser
    :return: parser
    """
    metadata_tables = [i2b2tablenames.concept_dimension,
                       i2b2tablenames.modifier_dimension,
                       i2b2tablenames.ontology_table,
                       i2b2tablenames.table_access]
    parser = FileAwareParser(description="FHIR in i2b2 metadata generator")
    # For reasons we don't completely understand, the default parser doesn't split the lines...
    parser.convert_arg_line_to_args = lambda arg_line: arg_line.split()
    parser.add_argument("-mv", "--metadatavoc", metavar="METAVOC URI",
                        help="Input directory or URI of w5.ttl and fhir.ttl files"
                             "(default: {})".format(Default_Metavoc_URI), default=Default_Metavoc_URI)
    parser.add_argument("-od", "--outdir", metavar="TSV OUTPUT DIR",
                        help="Output directory to store .tsv files. If absent, .tsv files are not generated.")
    parser.add_argument("-t", "--table", metavar="I2B2 TABLE",
                        help="Table to update ({}) (default: All tables)".format(', '.join(metadata_tables)),
                        choices=metadata_tables)
    parser.add_argument("-r", "--resource",
                        help="Name of specific resource to emit (e.g. Observation). (default: all)")
    parser.add_argument("--sourcesystem", metavar="SOURCESYSTEM_CD",
                        default=Default_Sourcesystem_Code,
                        help="sourcesystem code (default: \"{}\")".format(Default_Sourcesystem_Code))
    parser.add_argument("--base",
                        default=Default_Path_Base,
                        help="Concept dimension base path. (default: \"{}\")".format(Default_Path_Base))
    parser.add_argument("-l", "--load",
                        help="Load i2b2 SQL tables", action="store_true")
    parser.add_argument("-v", "--version", action='version', version='Version: {}'.format(__version__), default=None)
    parser.add_argument("--list", help="List table names", action="store_true")
    parser.add_argument("--test", help="Test the confguration", action="store_true")
    # Add the database connection arguments list
    add_connection_args(parser)

    return parser


def genargs(argv: List[str]) -> Namespace:
    """
    Generate the list of input arguments
    :param argv: input argument list
    :return: argparser namespace
    """

    def setdefault(self: Namespace, vn: str, default: object) -> None:
        assert vn in self, "Unknown option"
        if not getattr(self, vn):
            setattr(self, vn, default)

    parser = create_parser()
    opts = parser.parse_args(decode_file_args(argv))
    if not (opts.version or opts.list or opts.test or opts.load or opts.outdir):
        parser.print_help()
    opts.setdefault = lambda *a: setdefault(opts, *a)
    opts.updatedate = datetime.now()
    if not opts.metadatavoc.endswith(os.sep):
        opts.metadatavoc = os.path.join(opts.metadatavoc, '')
    if opts.outdir and not opts.outdir.endswith(os.sep):
        opts.outdir = os.path.join(opts.outdir, '')
    i2b2tablenames.ontology_table = opts.onttable
    # Set the defaults for the crc and ontology tables
    return process_parsed_args(opts) if opts.load or opts.list or opts.test else opts


def generate_i2b2(argv: List[str]) -> bool:
    """
    Generate a set of i2b2 metadata tables
    :param argv: input arguments
    :return: Success indicator
    """
    opts = genargs(argv)

    # configuration test option
    if opts.test:
        if not test_configuration(opts):
            return False

    # Load the ontology if actually generating output
    if opts.load or opts.outdir:
        g = load_fhir_ontology(opts)
    else:
        g = None

    # list table names
    opts.tables = I2B2Tables(opts) if (opts.load or opts.list) and not opts.test else None
    if opts.list:
        print('\n'.join(["{} : {}".format(tn, tp) for tn, tp in opts.tables._tables()]))

    # Generate actual output if requested
    if opts.load or opts.outdir:
        return g is not None and generate_i2b2_files(g, opts)
    else:
        return True


if __name__ == "__main__":
    generate_i2b2(sys.argv[1:])



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Optional, List, cast, Dict, NamedTuple, Set

from rdflib import URIRef, Graph

from i2fhirb2.fhir.fhirmetadata import FHIRMetadata
from i2fhirb2.fhir.fhirmetadatavocabulary import FMVGraphNode, FMVGraphEdge
from i2fhirb2.fhir.fhirspecific import concept_path, concept_path_sans_root, concept_code, rightmost_element, \
    skip_fhir_predicates, skip_fhir_types
from i2fhirb2.i2b2model.metadata.i2b2conceptdimension import ConceptDimension
from i2fhirb2.i2b2model.metadata.i2b2modifierdimension import ModifierDimension
from i2fhirb2.i2b2model.metadata.i2b2ontology import OntologyEntry, OntologyRoot, ConceptOntologyEntry, \
    ModifierOntologyEntry


class DimensionSet(NamedTuple):
    ontology_dimension: List[OntologyEntry]
    concept_dimension: List[ConceptDimension]
    modifier_dimension: List[ModifierDimension]


class FHIROntologyTable(FHIRMetadata):
    """  The set of i2b2 ontology table entries for the supplied subject or all root concepts
    """
    def __init__(self, g: Graph, name_base: str=None, modifier_base: str = None) -> None:
        super().__init__(g, name_base=name_base, modifier_base=modifier_base)
        OntologyEntry.graph = g

    @staticmethod
    def _all_edges(node: FMVGraphNode) -> Set[FMVGraphEdge]:
        """
        Follow the parents, pulling all edges
        :param node: target node
        :return: list of all edges back to root
        """
        rval = set([e for e in node.edges
                    if e.predicate not in skip_fhir_predicates and e.type_node.node not in skip_fhir_types])
        for p in node.parents:
            if p.node not in skip_fhir_types:
                rval.update(FHIROntologyTable._all_edges(p))
        return rval

    def _modifier_ontology_list(self, parent_path: str, parent_uri: URIRef, modifier_base_path: str, node: FMVGraphNode,
                                modifier_dims: Dict[str, ModifierDimension], in_multiple_elements: bool,
                                inside: Set[URIRef] = None, depth: int=1) \
            -> List[ModifierOntologyEntry]:
        """
        Return a list of modifier ontology entries for the supplied base and graph node
        :param parent_path: Path of parent node
        :param parent_uri: URI of base code
        :param modifier_base_path: Root of modifier path
        :param node: Graph node for URI
        :param in_multiple_elements:
        :param inside: Modifier types already being visited (to prevent recursion)
        :param depth: modifier depth
        :return:
        """
        if inside is None:
            inside = set()
        ontology_modifiers: List[ModifierOntologyEntry] = []
        for edge in self._all_edges(node):
            if edge.type_node.is_primitive or in_multiple_elements:
                modifier_path = self._modifier_base + modifier_base_path + rightmost_element(edge.predicate)[1:]
                ontology_modifiers.append(
                    ModifierOntologyEntry(depth,                # depth
                                          parent_uri,       # subject - uri of concept being modified
                                          edge.predicate,   # URI of the modifier itself
                                          modifier_path,    # path that the modifier is applied to
                                          parent_path,      # modifier path
                                          edge.type_node.is_primitive,  # is leaf
                                          edge.predicate,   #
                                          edge.type_node.node))  # actual type of modifier
                if edge.type_node.is_primitive:
                    modifier_entry = ModifierDimension(edge.predicate, self._name_base)
                    if modifier_entry.modifier_path not in modifier_dims:
                        modifier_dims[modifier_entry.modifier_path] = modifier_entry
                elif in_multiple_elements:
                    if edge.type_node not in inside:
                        inside.add(edge.type_node)
                        ontology_modifiers += \
                            self._modifier_ontology_list(parent_path,
                                                         parent_uri,
                                                         modifier_base_path + rightmost_element(edge.predicate)[1:],
                                                         edge.type_node,
                                                         modifier_dims,
                                                         True,
                                                         inside,
                                                         depth + 1)
                        inside.remove(edge.type_node)

        return ontology_modifiers

    def dimension_list(self, resource: Optional[URIRef]=None) -> DimensionSet:
        """
        Return the set of ontology entries for all w5 concepts and resources (or resource if it is supplied)
        :param resource: Optional resource URI -- for debugging purposes only.  None means all resources
        :return: List of i2b2 ontology entries
        """
        # TODO: Figure out how to get DomainResource and Resource concepts into this
        # TODO: DomainResource entries
        # TODO: Where is Observation.component???
        ontology_entries: List[OntologyEntry] = [cast(OntologyEntry, OntologyRoot(self._name_base))]    # FHIR root node
        concept_dimension_entries: Dict[str, ConceptDimension] = {}
        modifier_dimension_entries: Dict[str, ModifierDimension] = {}

        # Create an entry for all of the W5 paths
        for w5_node in self._fhir_metadatavocabulary.w5_graph.w5_paths():
            ontological_path = self._name_base + concept_path(w5_node.node)
            ontology_entries.append(
                ConceptOntologyEntry(w5_node.node,                  # subject URI
                                     w5_node.path,                  # navigational path
                                     ontological_path,              # Ontological path
                                     is_leaf=False,
                                     is_draggable=False))

            if w5_node.fhir_resource_uri and (resource is None or w5_node.fhir_resource_uri == resource):
                for conc_uri, conc_node_ent in self.fhir_concepts(w5_node.fhir_resource_uri).items():
                    conc_node = conc_node_ent.graph_node
                    if '.' in concept_code(conc_uri):
                        navigational_path = w5_node.path + concept_path_sans_root(conc_uri)
                        ontological_path = self._name_base + concept_path(conc_uri)
                        ontology_entries.append(
                            ConceptOntologyEntry(conc_uri,
                                                 navigational_path,
                                                 ontological_path,
                                                 is_leaf=conc_node.is_primitive,
                                                 is_draggable=True,
                                                 primitive_type=conc_node.node if conc_node.is_primitive else None))

                        concept_dimension_entries[ontological_path] = ConceptDimension(conc_uri, self._name_base)
                        ontology_entries += self._modifier_ontology_list(navigational_path,
                                                                         conc_uri,
                                                                         concept_path(conc_uri),
                                                                         conc_node,
                                                                         modifier_dimension_entries,
                                                                         conc_node_ent.is_multiple)

        return DimensionSet(ontology_entries,
                            list(concept_dimension_entries.values()),
                            list(modifier_dimension_entries.values()))



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional

from fhirtordf.rdfsupport.namespaces import W5, FHIR, namespace_for
from rdflib import Graph, URIRef
from rdflib.namespace import split_uri, RDFS
from rdflib.term import Node, BNode, Literal


# TODO: move this to a stable version when R4 is adopted

DEFAULT_FMV = "http://build.fhir.org/"
DEFAULT_PROVIDER_ID = "FHIR:DefaultProvider"

DEFAULT_NAME_BASE = '\\FHIR\\'              # Default hame base
DEFAULT_PROJECT_ID = 'fhir'                 # Default project id for mapping tables
DEFAULT_ENCOUNTER_NUMBER_START = 500000     # Default starting encounter number if none is present
DEFAULT_PATIENT_NUMBER_START = 100000001    # Default starting patient number if none is present
IDE_SOURCE_HIVE = "HIVE"                    # Source for number to id mapping

DEFAULT_ONTOLOGY_TABLE = "custom_meta"      # Default metadata ontology table

# List of W5 categores that should not be included in the output
w5_infrastructure_categories = {W5.conformance, W5.infrastructure, W5.information}

# List of predicates that aren't a direct part of the i2b2 structure
skip_fhir_predicates = {FHIR.index, FHIR.nodeRole, FHIR['id']}
skip_fhir_types = {FHIR.BackboneElement, FHIR.Element, FHIR.Resource}

# List of FHIR 'primitive' types, not to be further expanded
# TODO: FHIR.Reference shouldn't be in this list
fhir_primitives = {FHIR.Reference, FHIR.index, FHIR.nodeRole, FHIR.value}


def concept_path_sans_root(subject: URIRef) -> str:
    """
    Generate a concept path without the root node (e.g. Observation//component//code --> component//code
    :param subject: URI to trim
    :return: resulting path
    """
    return concept_path(subject).split('\\', 1)[1]


def concept_path(subject: URIRef) -> str:
    """
    Convert subject into an i2b2 concept path fragment.
    Example: Patient.status --> Patient\\status\\
    :param subject: FHIR URI
    :return: i2b2 path fragment
    """
    subj_path = split_uri(subject)[1]
    if is_w5_uri(subject):
        return (subj_path.rsplit('.', 1)[1] if '.' in subj_path else subj_path) + '\\'
    else:
        return split_uri(subject)[1].replace('.', '\\') + '\\'


def concept_code(subject: URIRef) -> str:
    """
    Return the i2b2 concept code for subject
    :param subject: URI to convert
    :return: 'ns:code' form of URI
    """
    ns, code = split_uri(subject)

    return '{}:{}'.format(namespace_for(ns).upper(), code)


def concept_name(g: Graph, subject: URIRef) -> str:
    """
    Return the i2b2 concept name for subject
    :param g: Graph - used to access label
    :param subject: concept subject
    :return: Name derived from label if it exists otherwise the URI itself
    """
    # Note - labels appear to have '.' in them as well
    return str(g.label(subject, split_uri(subject)[1])).replace('.', ' ')


def modifier_path(modifier: URIRef) -> str:
    """
    Convert modifier uri into an i2b2 modifier path fragment, removing the first part of the name
    Example: CodedEntry.code.text --> code\text\
    :param modifier: FHIR URI
    :return: i2b2 path fragment
    """
    path = split_uri(modifier)[1]
    return (path.split('.', 1)[1].replace('.', '\\') if '.' in path else path) + '\\'


def rightmost_element(uri: URIRef) -> str:
    """
    Isolate the rightmost element in a URI path.
    Example: CodedEntry.code.text --> \text\
             CodedEntry           --> \
    :param uri: input URI
    :return: rightmost element in path form
    """
    uri_path = split_uri(uri)[1]
    return '\\' + ((uri_path.rsplit('.', 1)[1] + '\\') if '.' in uri_path else "")


def modifier_code(modifier: URIRef) -> str:
    """
    Return the i2b2 modifier code for subject.  Output is same as concept_code
    :param modifier:
    :return:
    """
    return concept_code(modifier)


def modifier_name(g: Graph, modifier: URIRef) -> str:
    """
    Return the i2b2 concept name for modifier removing the first part of the name
    :param g: Graph - used to access label
    :param modifier: concept subject
    :return: Name derived from lable if it exists otherwise the URI itself
    """
    default_name = split_uri(modifier)[1]
    full_name = str(g.label(modifier, default_name))
    if '.' in full_name:
        full_name = default_name.split('.', 1)[1]        # Remove first name segment
    return full_name.replace('.', ' ')


def composite_uri(parent: URIRef, mod: URIRef) -> URIRef:
    """
    Return a composite URI consisting of the parent + '.' + the last element in the modifier
    :param parent: base URI
    :param mod: modifier URI
    :return: composite
    """
    p1 = split_uri(mod)[1]
    return URIRef(str(parent) + '.' + (p1.rsplit('.', 1)[1] if '.' in p1 else p1))


def is_w5_uri(uri: URIRef) -> bool:
    return split_uri(uri)[0] == str(W5)


def is_primitive(g: Graph, subj: Node) -> bool:
    """
    Determine whether subj is an instance of a FHIR primitive type
    :param g: Graph context for testing
    :param subj: element to test
    :return:
    """
    parents = set(g.objects(subj, RDFS.subClassOf))
    return FHIR.Primitive in parents or subj in fhir_primitives


def instance_is_primitive(g: Graph, obj: Optional[Node]) -> bool:
    """
    Determine whether obj is a FHIR 'primitive', which we can identify by testing for a fhir:value entry
    :param g: Graph context for testing
    :param obj: object to test
    :return:
    """
    if obj is None or not isinstance(obj, BNode):
        return False
    obj_vals = list(g.objects(obj, FHIR.value))
    return len(obj_vals) == 1 and isinstance(obj_vals[0], Literal)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
from argparse import ArgumentParser
import shlex
from typing import List

import sys

from i2fhirb2.loaders.i2b2graphmap import I2B2GraphMap
from i2fhirb2.sqlsupport.dbconnection import add_connection_args, process_parsed_args, I2B2Tables, decode_file_args, \
    FileAwareParser


def create_parser() -> FileAwareParser:
    """
    Create a command line parser
    :return: parser
    """
    parser = FileAwareParser(description="Clear data from FHIR observation fact table")
    parser.add_argument("uploadid", metavar="Upload identifiers",
                        help="Upload identifer(s) -- unique batch identifiers", type=int, nargs='*')
    parser.add_argument("--sourcesystemcd", metavar="Source system code",
                        help="sourcesystem_code to remove")
    parser.add_argument("-mv", "--metavoc", help="Metavocabulary directory - ignored")
    return parser


def remove_facts(argv: List[str]) -> bool:
    """
    Convert a set of FHIR resources into their corresponding i2b2 counterparts.
    :param argv: Command line arguments.  See: create_parser for details
    :return:
    """
    parser = add_connection_args(create_parser())
    opts = parser.parse_args(decode_file_args(argv, parser))
    if opts is None:
        return False
    process_parsed_args(opts)           # Update CRC and Meta table connection information
    if not opts.uploadid and not opts.sourcesystemcd:
        parser.print_usage()
        return False
    for uploadid in opts.uploadid:
        print("---> Removing entries for id {}".format(uploadid))
        I2B2GraphMap.clear_i2b2_tables(I2B2Tables(opts), uploadid)
    if opts.sourcesystemcd:
        print("---> Removing entries for sourcesystem_cd {}".format(opts.sourcesystemcd))
        I2B2GraphMap.clear_i2b2_sourcesystems(I2B2Tables(opts), opts.sourcesystemcd)
    return True


if __name__ == "__main__":
    remove_facts(sys.argv[1:])



from setuptools import setup

import sys

from i2fhirb2 import __version__

requires = ['SQLAlchemy', 'python_dateutil', 'rdflib', 'psycopg2', 'jsonasobj>=1.1.1',
            'dirlistproc', 'fhirtordf>=0.9.3', 'isodate']
if sys.version_info < (3, 5):
    requires.append('typing')

setup(
    name='i2FHIRb2',
    version=__version__,
    packages=['tests', 'scripts', 'i2fhirb2', 'i2fhirb2.fhir', 'i2fhirb2.i2b2model', 'i2fhirb2.sqlsupport'],
    url='https://github.com/BD2KOnFHIR/i2FHIRb2',
    license='Apache 2.0',
    author='Harold Solbrig',
    author_email='solbrig.harold@mayo.edu',
    description='FHIR in i2b2 model conversion tools',
    long_description='Toolkit to represent the "FHIR Ontology" in i2b2',
    install_requires=requires,
    scripts=['scripts/generate_i2b2', 'scripts/loadfacts', 'scripts/removefacts'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Healthcare Industry',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Database',
        'Programming Language :: Python :: 3 :: Only'
    ]
)
